from .errors import *
from .fetcher import *
from .url import *
from .utils import *
from .weverse import *

__title__ = 'Weverse.py'
__author__ = 'Anson Quek'
__license__ = 'MIT'
__version__ = '1.0.0.0'
